# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio import util


@click.command('raw')
@click.pass_context
@click.argument('cmd')
def cli(ctx, cmd):
    """Execute commands using Apio packages."""

    exit_code = util.call(cmd)
    ctx.exit(exit_code)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from os import listdir
from os.path import isfile
from sys import exit as sys_exit

from apio import util

commands_folder = util.get_folder('commands')


class ApioCLI(click.MultiCommand):

    def list_commands(self, ctx):
        rv = []
        for filename in listdir(commands_folder):
            if filename.startswith('__init__'):
                continue
            if filename.endswith('.py'):
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = util.safe_join(commands_folder, name + '.py')
        if isfile(fn):
            with open(fn) as f:
                code = compile(f.read(), fn, 'exec')
                eval(code, ns, ns)
            return ns.get('cli')


@click.command(cls=ApioCLI, invoke_without_command=True)
@click.pass_context
@click.version_option()
def cli(ctx):

    # Update help structure
    if ctx.invoked_subcommand is None:
        help = ctx.get_help()
        help = help.split('\n')

        setup_help = find_commands_help(help, ['drivers', 'init',
                                               'install', 'uninstall'])
        util_help = find_commands_help(help, ['boards', 'config', 'examples',
                                              'raw', 'system', 'upgrade'])

        help = '\n'.join(help)
        help = help.replace('Commands:\n', 'Project commands:\n')
        help += "\n\nSetup commands:\n"
        help += '\n'.join(setup_help)
        help += "\n\nUtility commands:\n"
        help += '\n'.join(util_help)

        click.secho(help)


def find_commands_help(help, commands):
    commands_help = []
    for line in list(help):
        for command in commands:
            if (' ' + command) in line:
                if line in help:
                    help.remove(line)
                    commands_help.append(line)
    return commands_help


if __name__ == '__main__':  # pragma: no cover
    sys_exit(cli())



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2
# -- Derived from:
# ---- Platformio project
# ---- (C) 2014-2016 Ivan Kravets <me@ikravets.com>
# ---- Licence Apache v2

import click
import requests

from email.utils import parsedate_tz
from math import ceil
from time import mktime

from apio import util

requests.packages.urllib3.disable_warnings()


class FDUnrecognizedStatusCode(util.ApioException):

    MESSAGE = 'Got an unrecognized status code \'{0}\' when downloaded {1}'


class FileDownloader(object):

    CHUNK_SIZE = 1024

    def __init__(self, url, dest_dir=None):
        self._url = url
        self._fname = url.split('/')[-1]

        self._destination = self._fname
        if dest_dir:
            self.set_destination(util.safe_join(dest_dir, self._fname))

        self._progressbar = None
        self._request = None

        # make connection
        self._request = requests.get(url, stream=True)
        if self._request.status_code != 200:
            raise FDUnrecognizedStatusCode(self._request.status_code, url)

    def set_destination(self, destination):
        self._destination = destination

    def get_filepath(self):
        return self._destination

    def get_lmtime(self):
        if 'last-modified' in self._request.headers:
            return self._request.headers.get('last-modified')

    def get_size(self):
        return int(self._request.headers.get('content-length'))

    def start(self):
        itercontent = self._request.iter_content(chunk_size=self.CHUNK_SIZE)
        f = open(self._destination, 'wb')
        chunks = int(ceil(self.get_size() / float(self.CHUNK_SIZE)))

        with click.progressbar(length=chunks, label='Downloading') as pb:
            for _ in pb:
                f.write(next(itercontent))
        f.close()
        self._request.close()

        self._preserve_filemtime(self.get_lmtime())

    def _preserve_filemtime(self, lmdate):
        if lmdate is not None:
            timedata = parsedate_tz(lmdate)
            lmtime = mktime(timedata[:9])
            util.change_filemtime(self._destination, lmtime)

    def __del__(self):
        if self._request:
            self._request.close()



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.managers.scons import SCons
from apio.managers.drivers import Drivers

# Python3 compat
import sys
if (sys.version_info > (3, 0)):
    unicode = str


@click.command('upload')
@click.pass_context
@click.option('-b', '--board', type=unicode, metavar='board',
              help='Set the board.')
@click.option('--serial-port', type=unicode, metavar='serial-port',
              help='Set the serial port.')
@click.option('--ftdi-id', type=unicode, metavar='ftdi-id',
              help='Set the FTDI id.')
@click.option('-s', '--sram', is_flag=True,
              help='Perform SRAM programming.')
@click.option('-p', '--project-dir', type=unicode, metavar='path',
              help='Set the target directory for the project.')
@click.option('-v', '--verbose', is_flag=True,
              help='Show the entire output of the command.')
@click.option('--verbose-yosys', is_flag=True,
              help='Show the yosys output of the command.')
@click.option('--verbose-arachne', is_flag=True,
              help='Show the arachne output of the command.')
def cli(ctx, board, serial_port, ftdi_id, sram, project_dir,
        verbose, verbose_yosys, verbose_arachne):
    """Upload the bitstream to the FPGA."""

    drivers = Drivers()
    drivers.pre_upload()
    # Run scons
    exit_code = SCons(project_dir).upload({
        'board': board,
        'verbose': {
            'all': verbose,
            'yosys': verbose_yosys,
            'arachne': verbose_arachne
        }
    }, serial_port, ftdi_id, sram)
    drivers.post_upload()
    ctx.exit(exit_code)

# Advanced notes: https://github.com/FPGAwars/apio/wiki/Commands#apio-upload



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import json
import click
import semantic_version

from os.path import isfile, isdir

from apio import util


class Profile(object):

    def __init__(self):
        self.config = {'exe': 'default', 'verbose': 0}
        self.labels = {'exe': 'Executable', 'verbose': 'Verbose'}
        self.settings = {}
        self.packages = {}
        self._profile_path = util.safe_join(
            util.get_home_dir(), 'profile.json')
        self.load()

    def installed_version(self, name, version):
        if name in self.packages:
            pkg_version = self.get_package_version(name)
            pkg_version = self._convert_old_version(pkg_version)
            version = self._convert_old_version(version)
            return (semantic_version.Version(pkg_version) ==
                    semantic_version.Version(version))

    def _convert_old_version(self, version):
        # Convert old versions to new format
        try:
            v = int(version)
            version = '1.{}.0'.format(v)
        except ValueError:
            pass
        return version

    def check_exe_default(self):
        return self.config.get('exe', '') == 'default'

    def add_package(self, name, version):
        self.packages[name] = {'version': version}

    def add_setting(self, key, value):
        self.settings[key] = value

    def add_config(self, key, value):
        if self.config.get(key, None) != value:
            self.config[key] = value
            self.save()
            click.secho('{0} mode updated: {1}'.format(
                self.labels.get(key, ''), value), fg='green')
        else:
            click.secho('{0} mode already {1}'.format(
                self.labels.get(key, ''), value), fg='yellow')

    def remove_package(self, name):
        if name in self.packages.keys():
            del self.packages[name]

    def get_verbose_mode(self):
        return int(self.config.get('verbose', False))

    def get_package_version(self, name, release_name=''):
        version = '0.0.0'
        if name in self.packages:
            version = self.packages.get(name).get('version')
        elif release_name:
            dir_name = util.get_package_dir(release_name)
            if isdir(dir_name):
                filepath = util.safe_join(dir_name, 'package.json')
                try:
                    with open(filepath, 'r') as json_file:
                        tmp_data = json.load(json_file)
                        if 'version' in tmp_data.keys():
                            version = tmp_data.get('version')
                except Exception:
                    pass
        return version

    def load(self):
        if isfile(self._profile_path):
            with open(self._profile_path, 'r') as profile:
                try:
                    self._load_profile(profile)
                except Exception:
                    pass

    def _load_profile(self, profile):
        data = json.load(profile)
        if 'config' in data.keys():
            self.config = data.get('config')
            if 'exe' not in self.config.keys():
                self.config['exe'] = 'default'
            if 'verbose' not in self.config.keys():
                self.config['verbose'] = 0
        if 'settings' in data.keys():
            self.settings = data.get('settings')
        if 'packages' in data.keys():
            self.packages = data.get('packages')
        else:
            self.packages = data  # Backward compatibility

    def save(self):
        util.mkdir(self._profile_path)
        with open(self._profile_path, 'w') as profile:
            data = {
                'config': self.config,
                'settings': self.settings,
                'packages': self.packages
            }
            json.dump(data, profile, indent=4, sort_keys=True)

    def list(self):
        for key in self.config:
            click.secho('{0} mode: {1}'.format(
                    self.labels.get(key, ''), self.config.get(key, '')),
                fg='yellow')



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.managers.installer import Installer
from apio.resources import Resources

platforms = ['linux_x86_64',
             'linux_i686',
             'linux_armv7l',
             'linux_aarch64',
             'windows',
             'darwin']


@click.command('uninstall')
@click.pass_context
@click.argument('packages', nargs=-1)
@click.option('-a', '--all', is_flag=True,
              help='Uninstall all packages.')
@click.option('-l', '--list', is_flag=True,
              help='List all installed packages.')
@click.option('-p', '--platform', type=click.Choice(platforms),
              metavar='platform',
              help='Set the platform [{}] (Advanced).'.format(
                ', '.join(platforms)))
def cli(ctx, packages, all, list, platform):
    """Uninstall packages."""

    if packages:
        _uninstall(packages, platform)
    elif all:  # pragma: no cover
        packages = Resources(platform).packages
        _uninstall(packages, platform)
    elif list:
        Resources(platform).list_packages(installed=True, notinstalled=False)
    else:
        click.secho(ctx.get_help())


def _uninstall(packages, platform):
    if click.confirm('Do you want to continue?'):
        for package in packages:
            Installer(package, platform, checkversion=False).uninstall()
    else:
        click.secho('Abort!', fg='red')



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2
# -- Derived from:
# ---- Platformio project
# ---- (C) 2014-2016 Ivan Kravets <me@ikravets.com>
# ---- Licence Apache v2

import os
import re
import sys
import jwt
import json
import click
import locale
import platform
import subprocess
import semantic_version
from threading import Thread
from os.path import expanduser, isdir, isfile, join, dirname, exists, normpath

from apio import LOAD_CONFIG_DATA

import requests
requests.packages.urllib3.disable_warnings()

# Python3 compat
if (sys.version_info > (3, 0)):
    unicode = str


class ApioException(Exception):

    MESSAGE = None

    def __str__(self):  # pragma: no cover
        if self.MESSAGE:
            return self.MESSAGE.format(*self.args)
        else:
            return Exception.__str__(self)


class AsyncPipe(Thread):  # pragma: no cover

    def __init__(self, outcallback=None):
        Thread.__init__(self)
        self.outcallback = outcallback

        self._fd_read, self._fd_write = os.pipe()
        self._pipe_reader = os.fdopen(self._fd_read)
        self._buffer = []

        self.start()

    def get_buffer(self):
        return self._buffer

    def fileno(self):
        return self._fd_write

    def run(self):
        for line in iter(self._pipe_reader.readline, ''):
            line = line.strip()
            self._buffer.append(line)
            if self.outcallback:
                self.outcallback(line)
        self._pipe_reader.close()

    def close(self):
        os.close(self._fd_write)
        self.join()


def get_systype():
    type_ = platform.system().lower()
    arch = platform.machine().lower()
    if type_ == 'windows':
        arch = 'amd64' if platform.architecture()[0] == '64bit' else 'x86'
    return '%s_%s' % (type_, arch) if arch else type_


try:
    codepage = locale.getdefaultlocale()[1]
    if 'darwin' in get_systype():
        UTF = True
    else:
        UTF = codepage.lower().find('utf') >= 0
except Exception:
    # Incorrect locale implementation, assume the worst
    UTF = False


def unicoder(p):
    """ Make sure a Unicode string is returned """
    if isinstance(p, unicode):
        return p
    if isinstance(p, str):
        return decoder(p)
    else:
        return unicode(decoder(p))


def decoder(p):
    if UTF:
        try:
            return p.decode('utf-8')
        except Exception:
            return p.decode(codepage)
    return p.decode(codepage)


def safe_join(*paths):
    """ Join paths in a Unicode-safe way """
    try:
        return join(*paths)
    except UnicodeDecodeError:
        npaths = ()
        for path in paths:
            npaths += (unicoder(path),)
        return join(*npaths)


def _get_config_data():
    config_data = None
    if LOAD_CONFIG_DATA:
        filepath = safe_join(os.sep, 'etc', 'apio.json')
        if isfile(filepath):
            with open(filepath, 'r') as f:
                # Load the JSON file
                config_data = json.loads(f.read())
    return config_data


config_data = _get_config_data()


def _get_projconf_option_dir(name, default=None):
    _env_name = 'APIO_%s' % name.upper()
    if _env_name in os.environ:
        return os.getenv(_env_name)
    if config_data and _env_name in config_data.keys():
        return config_data.get(_env_name)
    return default


def get_home_dir():
    home_dir = _get_projconf_option_dir('home_dir', '~/.apio')
    home_dir = re.sub(r'\~', expanduser('~').replace('\\', '/'), home_dir)

    paths = home_dir.split(os.pathsep)
    path = _check_writable(paths)
    if not path:
        path = _create_path(paths)
        if not path:
            click.secho('Error: no usable home directory ' + path, fg='red')
            exit(1)
    return path


def _check_writable(paths):
    ret = ''
    for path in paths:
        if isdir(path):
            if os.access(path, os.W_OK):
                # Path is writable
                ret = path
                break
            else:
                click.secho('Warning: can\'t write in path ' + path,
                            fg='yellow')
    return ret


def _create_path(paths):
    ret = ''
    for path in paths:
        if not isdir(path):
            try:
                os.makedirs(path)
                ret = path
                break
            except OSError as ioex:
                if ioex.errno == 13:
                    click.secho('Warning: can\'t create ' + path,
                                fg='yellow')
    return ret


def get_package_dir(pkg_name):
    home_dir = _get_projconf_option_dir('pkg_dir', '')
    if not home_dir:
        home_dir = _get_projconf_option_dir('home_dir', '~/.apio')
    home_dir = re.sub(r'\~', expanduser('~').replace('\\', '/'), home_dir)

    paths = home_dir.split(os.pathsep)
    for path in paths:
        package_dir = safe_join(path, 'packages', pkg_name)
        if isdir(package_dir):
            return package_dir

    return ''


def get_project_dir():
    return os.getcwd()


def setup_environment():
    base_dirs = {
        'scons': get_package_dir('tool-scons'),
        'system': get_package_dir('tools-system'),
        'icestorm': get_package_dir('toolchain-icestorm'),
        'iverilog': get_package_dir('toolchain-iverilog'),
        'verilator': get_package_dir('toolchain-verilator'),
        'gtkwave': get_package_dir('tool-gtkwave')
    }
    bin_dirs = {
        'scons': safe_join(base_dirs.get('scons'), 'script'),
        'system': safe_join(base_dirs.get('system'), 'bin'),
        'icestorm': safe_join(base_dirs.get('icestorm'), 'bin'),
        'iverilog': safe_join(base_dirs.get('iverilog'), 'bin'),
        'verilator': safe_join(base_dirs.get('verilator'), 'bin'),
        'gtkwave': safe_join(base_dirs.get('gtkwave'), 'bin')
    }

    # Give the priority to the python packages installed with apio
    os.environ['PATH'] = os.pathsep.join([
        get_bin_dir(),
        os.environ['PATH']
    ])

    # Give the priority to the packages installed by apio
    os.environ['PATH'] = os.pathsep.join([
        bin_dirs.get('system'),
        bin_dirs.get('icestorm'),
        bin_dirs.get('iverilog'),
        bin_dirs.get('verilator'),
        os.environ['PATH']
    ])

    if platform.system() == 'Windows':
        os.environ['PATH'] = os.pathsep.join([
            bin_dirs.get('gtkwave'),
            os.environ['PATH']
        ])

    # Add environment variables
    if not config_data:  # /etc/apio.json file does not exist
        os.environ['IVL'] = safe_join(
            base_dirs.get('iverilog'), 'lib', 'ivl')
    os.environ['VLIB'] = safe_join(
        base_dirs.get('iverilog'), 'vlib')
    os.environ['ICEBOX'] = safe_join(
        base_dirs.get('icestorm'), 'share', 'icebox')
    os.environ['VERLIB'] = safe_join(
        base_dirs.get('verilator'), 'share')

    return bin_dirs


def resolve_packages(packages, installed_packages, spec_packages):
    bin_dirs = setup_environment()

    # -- Check packages
    check = True
    for package in packages:
        version = installed_packages.get(package, {}).get('version', '')
        spec_version = spec_packages.get(package, '')
        check &= check_package(
            package,
            version,
            spec_version,
            bin_dirs.get(package))

    # -- Load packages
    if check:
        global scons_command
        scons_command = [normpath(sys.executable),
                         safe_join(bin_dirs.get('scons'), 'scons')]

    return check


def check_package(name, version, spec_version, path):
    # Apio package 'gtkwave' only exists for Windows.
    # Linux and MacOS user must install the native GTKWave.
    if name == 'gtkwave' and platform.system() != 'Windows':
        return True

    # Check package path
    if not isdir(path):
        show_package_path_error(name)
        show_package_install_instructions(name)
        return False

    # Check package version
    if not check_package_version(version, spec_version):
        show_package_version_warning(name, version, spec_version)
        show_package_install_instructions(name)
        return False

    return True


def check_package_version(version, spec_version):
    try:
        spec = semantic_version.Spec(spec_version)
        return semantic_version.Version(version) in spec
    except ValueError:
        pass


def show_package_version_warning(name, version, spec_version):
    message = ('Warning: package \'{0}\' version {1}\n'
               'does not match the semantic version {2}').format(
        name, version, spec_version)
    click.secho(message, fg='yellow')


def show_package_path_error(name):
    message = 'Error: package \'{}\' is not installed'.format(name)
    click.secho(message, fg='red')


def show_package_install_instructions(name):
    if config_data and _check_apt_get():  # /etc/apio.json file exists
        click.secho('Please run:\n'
                    '   apt-get install apio-{}'.format(name), fg='yellow')
    else:
        click.secho('Please run:\n'
                    '   apio install {}'.format(name), fg='yellow')


def _check_apt_get():
    """Check if apio can be installed through apt-get"""
    check = False
    if 'TESTING' not in os.environ:
        result = exec_command(['dpkg', '-l', 'apio'])
        if result and result.get('returncode') == 0:
            match = re.findall('rc\s+apio', result.get('out')) + \
                    re.findall('ii\s+apio', result.get('out'))
            check = len(match) > 0
    return check


def get_package_version(name, profile):
    version = ''
    if name in profile.packages:
        version = profile.packages.get(name).get('version')
    return version


def get_package_spec_version(name, resources):
    spec_version = ''
    if name in resources.distribution.get('packages'):
        spec_version = resources.distribution.get('packages').get(name)
    return spec_version


def change_filemtime(path, time):
    os.utime(path, (time, time))


def exec_command(*args, **kwargs):  # pragma: no cover
    result = {
        'out': None,
        'err': None,
        'returncode': None
    }

    default = dict(
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=platform.system() == 'Windows'
    )
    default.update(kwargs)
    kwargs = default

    try:
        p = subprocess.Popen(*args, **kwargs)
        result['out'], result['err'] = p.communicate()
        result['returncode'] = p.returncode
    except KeyboardInterrupt:
        click.secho('Aborted by user', fg='red')
        exit(1)
    except Exception as e:
        click.secho(str(e), fg='red')
        exit(1)
    finally:
        for s in ('stdout', 'stderr'):
            if isinstance(kwargs[s], AsyncPipe):
                kwargs[s].close()

    _parse_result(kwargs, result)

    return result


def _parse_result(kwargs, result):
    for s in ('stdout', 'stderr'):
        if isinstance(kwargs[s], AsyncPipe):
            result[s[3:]] = '\n'.join(kwargs[s].get_buffer())

    for k, v in result.items():
        if v and isinstance(v, unicode):
            result[k].strip()


def call(cmd):
    setup_environment()
    return subprocess.call(cmd, shell=True)


def get_pypi_latest_version():
    r = None
    version = None
    try:
        r = requests.get('https://pypi.python.org/pypi/apio/json')
        version = r.json().get('info').get('version')
        r.raise_for_status()
    except requests.exceptions.ConnectionError as e:
        error_message = str(e)
        if 'NewConnectionError' in error_message:
            click.secho('Error: could not connect to Pypi.\n'
                        'Check your internet connection and try again',
                        fg='red')
        else:
            click.secho(error_message, fg='red')
    except Exception as e:
        click.secho('Error: ' + str(e), fg='red')
    finally:
        if r:
            r.close()
    return version


def get_folder(folder):
    return safe_join(dirname(__file__), folder)


def mkdir(path):
    path = dirname(path)
    if not exists(path):
        try:
            os.makedirs(path)
        except OSError:
            pass


def check_dir(_dir):
    if _dir is None:
        _dir = os.getcwd()

    if isfile(_dir):
        click.secho(
            'Error: project directory is already a file: {0}'.format(_dir),
            fg='red')
        exit(1)

    if not exists(_dir):
        try:
            os.makedirs(_dir)
        except OSError:
            pass
    return _dir


def command(function):
    """Command decorator"""
    def decorate(*args, **kwargs):
        exit_code = 1
        try:
            exit_code = function(*args, **kwargs)
        except Exception as e:
            if str(e):
                click.secho('Error: ' + str(e), fg='red')
        finally:
            return exit_code
    return decorate


def decode(text):
    return jwt.decode(text, 'secret', algorithm='HS256')


def get_serial_ports():
    from serial.tools.list_ports import comports
    result = []

    for port, description, hwid in comports():
        if not port:
            continue
        if 'VID:PID' in hwid:
            result.append({
                'port': port,
                'description': description,
                'hwid': hwid
            })

    return result


def get_tinyprog_meta():
    command = join(get_bin_dir(), 'tinyprog')
    result = exec_command([command, '--pyserial', '--meta'])
    try:
        out = unicoder(result.get('out', ''))
        if out:
            return json.loads(out)
    except Exception as e:
        print(e)
        return []
    return []


def get_bin_dir():
    candidate = dirname(sys.modules['__main__'].__file__)
    # Windows + virtualenv = 💩
    # In this case the main file is: venv/Scripts/apio.exe/__main__.py!
    # This is not good because venv/Scripts/apio.exe is not a directory
    # So here we go with the workaround:
    if candidate.endswith('.exe'):
        return dirname(candidate)
    else:
        return candidate


def get_python_version():
    return '{0}.{1}'.format(sys.version_info[0], sys.version_info[1])



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo, Juan González
# -- Licence GPLv2

import os
import glob
import click
import codecs
import shutil

from os.path import sep, isdir, isfile, dirname, basename

from apio import util
from apio.profile import Profile
from apio.resources import Resources

# -- Error messages
EXAMPLE_NOT_FOUND_MSG = """
Warning: this example does not exist
Use `apio examples -l` for listing all the available examples"""

EXAMPLE_OF_USE_CAD = """
Example of use:
   apio examples -f leds
Copy the leds example files to the current directory"""

EXAMPLE_DIR_FILE = """
To get an example, use the command:
   apio examples -d/-f name"""


class Examples(object):

    def __init__(self):
        profile = Profile()
        resources = Resources()

        self.name = 'examples'
        self.examples_dir = util.get_package_dir(self.name)
        self.version = util.get_package_version(self.name, profile)
        self.spec_version = util.get_package_spec_version(self.name, resources)

    def list_examples(self):
        if util.check_package(
            self.name,
            self.version,
            self.spec_version,
            self.examples_dir
        ):
            # examples = sorted(os.listdir(self.examples_dir))
            examples = [dirname(y).replace(self.examples_dir + sep, '')
                        for x in os.walk(self.examples_dir)
                        for y in glob.glob(util.safe_join(x[0], 'info'))]
            click.secho('')
            for example in examples:
                example_dir = util.safe_join(self.examples_dir, example)
                if isdir(example_dir):
                    info_path = util.safe_join(example_dir, 'info')
                    info = ''
                    if isfile(info_path):
                        with codecs.open(info_path, 'r', 'utf-8') as info_file:
                            info = info_file.read().replace('\n', '')
                    click.secho(' ' + example, fg='blue', bold=True)
                    click.secho('-' * click.get_terminal_size()[0])
                    click.secho(' ' + info)
                    click.secho('')
            click.secho(EXAMPLE_DIR_FILE, fg='green')
            click.secho(EXAMPLE_OF_USE_CAD, fg='green')
        else:
            return 1
        return 0

    def copy_example_dir(self, example, project_dir, sayno):
        if util.check_package(
            self.name,
            self.version,
            self.spec_version,
            self.examples_dir
        ):
            project_dir = util.check_dir(project_dir)
            example_path = util.safe_join(project_dir, example)
            local_example_path = util.safe_join(self.examples_dir, example)

            if isdir(local_example_path):
                if isdir(example_path):

                    # -- If sayno, do not copy anything
                    if not sayno:
                        click.secho(
                            'Warning: ' + example +
                            ' directory already exists', fg='yellow')

                        if click.confirm('Do you want to replace it?'):
                            shutil.rmtree(example_path)
                            self._copy_dir(example, local_example_path,
                                           example_path)
                elif isfile(example_path):
                    click.secho(
                        'Warning: ' + example + ' is already a file',
                        fg='yellow')
                else:
                    self._copy_dir(example, local_example_path, example_path)
            else:
                click.secho(EXAMPLE_NOT_FOUND_MSG, fg='yellow')
        else:
            return 1
        return 0

    def copy_example_files(self, example, project_dir, sayno):
        if util.check_package(
            self.name,
            self.version,
            self.spec_version,
            self.examples_dir
        ):
            project_dir = util.check_dir(project_dir)
            example_path = project_dir
            local_example_path = util.safe_join(self.examples_dir, example)

            if isdir(local_example_path):
                self._copy_files(example, local_example_path,
                                 example_path, sayno)
            else:
                click.secho(EXAMPLE_NOT_FOUND_MSG, fg='yellow')
        else:
            return 1
        return 0

    def _copy_files(self, example, src_path, dest_path, sayno):
        click.secho('Copying ' + example + ' example files ...')
        example_files = glob.glob(util.safe_join(src_path, '*'))
        for f in example_files:
            filename = basename(f)
            if filename != 'info':
                filepath = util.safe_join(dest_path, filename)
                if isfile(filepath):

                    # -- If sayno, do not copy the file. Move to the next
                    if sayno:
                        continue

                    click.secho(
                        'Warning: ' + filename + ' file already exists',
                        fg='yellow')
                    if click.confirm('Do you want to replace it?'):
                        shutil.copy(f, dest_path)
                elif isdir(filepath):
                    click.secho(
                        'Warning: ' + filename + ' is already a directory',
                        fg='yellow')
                    return
                else:
                    shutil.copy(f, filepath)
        click.secho(
            'Example files \'{}\' have been successfully created!'.format(
                example
            ), fg='green')

    def _copy_dir(self, example, src_path, dest_path):
        click.secho('Creating ' + example + ' directory ...')
        shutil.copytree(src_path, dest_path)
        click.secho(
            'Example \'' + example + '\' has been successfully created!',
            fg='green')

    def examples_of_use_cad(self):
        return EXAMPLE_OF_USE_CAD



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Authors Jesús Arroyo, Miodrag Milanovic
# -- Licence GPLv2

import click

from apio.managers.scons import SCons

# Python3 compat
import sys
if (sys.version_info > (3, 0)):
    unicode = str


@click.command('lint')
@click.pass_context
@click.option('-a', '--all', is_flag=True,
              help='Enable all warnings, including code style warnings.')
@click.option('-t', '--top', type=unicode, metavar='top',
              help='Set top module.')
@click.option('--nostyle', is_flag=True,
              help='Disable all style warnings.')
@click.option('--nowarn', type=unicode, metavar='nowarn',
              help='Disable specific warning(s).')
@click.option('--warn', type=unicode, metavar='warn',
              help='Enable specific warning(s).')
@click.option('-p', '--project-dir', type=unicode, metavar='path',
              help='Set the target directory for the project.')
def cli(ctx, all, top, nostyle, nowarn, warn, project_dir):
    """Lint the verilog code."""

    exit_code = SCons(project_dir).lint({
        'all': all,
        'top': top,
        'nostyle': nostyle,
        'nowarn': nowarn,
        'warn': warn
    })
    ctx.exit(exit_code)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.resources import Resources


@click.command('boards')
@click.pass_context
@click.option('-l', '--list', is_flag=True,
              help='List all supported FPGA boards.')
@click.option('-f', '--fpga', is_flag=True,
              help='List all supported FPGA chips.')
def cli(ctx, list, fpga):
    """Manage FPGA boards."""

    if list:
        Resources().list_boards()
    elif fpga:
        Resources().list_fpgas()
    else:
        click.secho(ctx.get_help())



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.managers.drivers import Drivers


@click.command('drivers')
@click.pass_context
@click.option('--ftdi-enable', is_flag=True,
              help='Enable FTDI drivers.')
@click.option('--ftdi-disable', is_flag=True,
              help='Disable FTDI drivers.')
@click.option('--serial-enable', is_flag=True,
              help='Enable Serial drivers.')
@click.option('--serial-disable', is_flag=True,
              help='Disable Serial drivers.')
def cli(ctx, ftdi_enable, ftdi_disable, serial_enable, serial_disable):
    """Manage FPGA boards drivers."""

    exit_code = 0

    if ftdi_enable:   # pragma: no cover
        exit_code = Drivers().ftdi_enable()
    elif ftdi_disable:   # pragma: no cover
        exit_code = Drivers().ftdi_disable()
    elif serial_enable:   # pragma: no cover
        exit_code = Drivers().serial_enable()
    elif serial_disable:   # pragma: no cover
        exit_code = Drivers().serial_disable()
    else:
        click.secho(ctx.get_help())

    ctx.exit(exit_code)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Juan González, Jesús Arroyo
# -- Licence GPLv2

import sys
import json
import click

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

from os.path import isfile

from apio import util
from apio.resources import Resources

PROJECT_FILENAME = 'apio.ini'


class Project(object):

    def __init__(self):
        self.board = None

    def create_sconstruct(self, project_dir='', sayyes=False):
        """Creates a default SConstruct file"""

        project_dir = util.check_dir(project_dir)

        sconstruct_name = 'SConstruct'
        sconstruct_path = util.safe_join(project_dir, sconstruct_name)
        local_sconstruct_path = util.safe_join(
            util.get_folder('resources'), sconstruct_name)

        if isfile(sconstruct_path):
            # -- If sayyes, skip the question
            if sayyes:
                self._copy_sconstruct_file(sconstruct_name, sconstruct_path,
                                           local_sconstruct_path)
            else:
                click.secho(
                    'Warning: {} file already exists'.format(sconstruct_name),
                    fg='yellow')

                if click.confirm('Do you want to replace it?'):
                    self._copy_sconstruct_file(sconstruct_name,
                                               sconstruct_path,
                                               local_sconstruct_path)
                else:
                    click.secho('Abort!', fg='red')

        else:
            self._copy_sconstruct_file(sconstruct_name, sconstruct_path,
                                       local_sconstruct_path)

    def create_ini(self, board, project_dir='', sayyes=False):
        """Creates a new apio project file"""

        project_dir = util.check_dir(project_dir)

        ini_path = util.safe_join(project_dir, PROJECT_FILENAME)

        # Check board
        boards = Resources().boards
        if board not in boards.keys():
            click.secho(
                'Error: no such board \'{}\''.format(board),
                fg='red')
            sys.exit(1)

        if isfile(ini_path):
            # -- If sayyes, skip the question
            if sayyes:
                self._create_ini_file(board, ini_path, PROJECT_FILENAME)
            else:
                click.secho(
                    'Warning: {} file already exists'.format(PROJECT_FILENAME),
                    fg='yellow')
                if click.confirm('Do you want to replace it?'):
                    self._create_ini_file(board, ini_path, PROJECT_FILENAME)
                else:
                    click.secho('Abort!', fg='red')
        else:
            self._create_ini_file(board, ini_path, PROJECT_FILENAME)

    def _create_ini_file(self, board, ini_path, ini_name):
        click.secho('Creating {} file ...'.format(ini_name))
        with open(ini_path, 'w') as file:
            config = ConfigParser.ConfigParser()
            config.add_section('env')
            config.set('env', 'board', board)
            config.write(file)
            click.secho(
                'File \'{}\' has been successfully created!'.format(
                    ini_name),
                fg='green')

    def _copy_sconstruct_file(self, sconstruct_name,
                              sconstruct_path, local_sconstruct_path):
        click.secho('Creating {} file ...'.format(sconstruct_name))
        with open(sconstruct_path, 'w') as sconstruct:
            with open(local_sconstruct_path, 'r') as local_sconstruct:
                sconstruct.write(local_sconstruct.read())
                click.secho(
                    'File \'{}\' has been successfully created!'.format(
                        sconstruct_name),
                    fg='green')

    def read(self):
        """Read the project config file"""

        # -- If no project finel found, just return
        if not isfile(PROJECT_FILENAME):
            print('Info: No {} file'.format(PROJECT_FILENAME))
            return

        # -- Read stored board
        board = self._read_board()

        # -- Update board
        self.board = board
        if not board:
            print('Error: invalid {} project file'.format(
                PROJECT_FILENAME))
            print('No \'board\' field defined in project file')
            sys.exit(1)

    def _read_board(self):
        board = ''

        # -- Read config file: old JSON format
        with open(PROJECT_FILENAME, 'r') as f:
            try:
                data = json.loads(f.read())
                board = data.get('board')
            except Exception:
                pass

        # -- Read config file: new CFG format
        if board == '':
            try:
                config = ConfigParser.ConfigParser()
                config.read(PROJECT_FILENAME)
                board = config.get('env', 'board')
            except Exception:
                print('Error: invalid {} project file'.format(
                    PROJECT_FILENAME))
                sys.exit(1)

        return board



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.profile import Profile


@click.command('config')
@click.pass_context
@click.option('-l', '--list', is_flag=True,
              help='List all configuration parameters.')
@click.option('-v', '--verbose', type=click.Choice(['0', '1']),
              help='Verbose mode: `0` General, `1` Information.')
@click.option('-e', '--exe', type=click.Choice(['default', 'native']),
              help='Configure executables: `default` selects apio packages, ' +
                   '`native` selects system binaries.')
def cli(ctx, list, verbose, exe):
    """Apio configuration."""

    if list:  # pragma: no cover
        profile = Profile()
        profile.list()
    elif verbose:  # pragma: no cover
        profile = Profile()
        profile.add_config('verbose', verbose)
    elif exe:  # pragma: no cover
        profile = Profile()
        profile.add_config('exe', exe)
    else:
        click.secho(ctx.get_help())
